# Maze Runner HTML
import sys, os, shutil

# Maze file constants:
WALL = '#'
EMPTY = ' '
START = 'S'
EXIT = 'E'

NORTH = 'NORTH'
SOUTH = 'SOUTH'
EAST = 'EAST'
WEST = 'WEST'

print('''MAZE RUNNER HTML
By Al Sweigart al@inventwithpython.com

(Maze files are generated by mazemaker*.py)''')

# Get the maze file's filename from the user:
while True:
    print('Enter the filename of the maze (or "quit"):')
    filename = input()

    if filename.upper() == 'QUIT':
        sys.exit()

    if os.path.exists(filename):
        break
    print('There is no file named', filename)

# Get the folder name to save the files to:
while True:
    print('Enter the folder to save the maze files to:')
    outputFolderName = input()

    if os.path.exists(outputFolderName):
        if os.path.isfile(outputFolderName):
            print(f'A file named {outputFolderName} already exists.')
            continue
        else:
            print(f'A folder named {outputFolderName} already exists. Use this name anyway? Y/N')
            useAnyway = input().upper()
            if not useAnyway.startswith('Y'):
                continue
    break

# Load the maze from a file:
mazeFile = open(filename)
maze = {}
lines = mazeFile.readlines()
startx = None
starty = None
exitx = None
exity = None
y = 0
for line in lines:
    WIDTH = len(line.rstrip())
    for x, character in enumerate(line.rstrip()):
        assert character in (WALL, EMPTY, START, EXIT), f'Invalid character at column {x + 1}, line {y + 1}'
        if character in (WALL, EMPTY):
            maze[(x, y)] = character
        elif character == START:
            startx, starty = x, y
            maze[(x, y)] = EMPTY
        elif character == EXIT:
            exitx, exity = x, y
            maze[(x, y)] = EMPTY
    y += 1
HEIGHT = y

assert startx != None and starty != None, 'Missing start point in maze file.'
assert exitx != None and exity != None, 'Missing exit point in maze file.'


# Generate HTML files:

# TODO, file format is: "3_12_NORTH.html"
numFilesWritten = 0
os.makedirs(outputFolderName, exist_ok=True)
for x in range(WIDTH):
    for y in range(HEIGHT):
        # If this space is a wall, then the player can't be here, so skip this file.
        if maze[(x, y)] == WALL:
            continue

        for direction in (NORTH, EAST, SOUTH, WEST):
            htmlFilename = f'{outputFolderName}/{x}_{y}_{direction}.html'
            print(f'Writing {htmlFilename}...')
            numFilesWritten += 1

            with open(htmlFilename, 'w') as htmlFile:
                if direction == NORTH:
                    # Map of the sections, relative  A
                    # to the player @:              BCD (Player facing north)
                    #                               E@F
                    offsets = (('A', 0, -2), ('B', -1, -1), ('C', 0, -1), ('D', 1, -1), ('E', -1, 0), ('F', 1, 0))
                if direction == SOUTH:
                    # Map of the sections, relative F@E
                    # to the player @:              DCB (Player facing south)
                    #                                A
                    offsets = (('A', 0, 2), ('B', 1, 1), ('C', 0, 1), ('D', -1, 1), ('E', 1, 0), ('F', -1, 0))
                if direction == EAST:
                    # Map of the sections, relative EB
                    # to the player @:              @CA (Player facing east)
                    #                               FD
                    offsets = (('A', 2, 0), ('B', 1, -1), ('C', 1, 0), ('D', 1, 1), ('E', 0, -1), ('F', 0, 1))
                if direction == WEST:
                    # Map of the sections, relative  DF
                    # to the player @:              AC@ (Player facing west)
                    #                                BE
                    offsets = (('A', -2, 0), ('B', -1, 1), ('C', -1, 0), ('D', -1, -1), ('E', 0, 1), ('F', 0, -1))

                section = {}
                for sec, xOffset, yOffset in offsets:
                    section[sec] = maze.get((x + xOffset, y + yOffset), WALL)
                    if (x + xOffset, y + yOffset) == (exitx, exity):
                        section[sec] = EXIT

                wallImageFilename = ''
                for sec in 'ABDCEF':
                    if section[sec] == WALL:
                        wallImageFilename += sec
                wallImageFilename = ''.join(sorted(list(wallImageFilename))) # put this back into alphabetical order because C and D are out of order

                # Draw the EXIT sign if needed:
                if section['C'] == EXIT:
                    wallImageFilename += '_exitback'
                if section['E'] == EXIT:
                    wallImageFilename += '_exitleft'
                if section['F'] == EXIT:
                    wallImageFilename += '_exitright'

                wallImageFilename += '.jpg'

                # Calculate turn URLs:
                leftDirection = {NORTH: WEST, WEST: SOUTH, SOUTH: EAST, EAST: NORTH}[direction]
                turnLeftURL = f'{x}_{y}_{leftDirection}.html'

                rightDirection = {NORTH: EAST, EAST: SOUTH, SOUTH: WEST, WEST: NORTH}[direction]
                turnRightURL = f'{x}_{y}_{rightDirection}.html'

                if section['C'] != WALL:
                    forwardURL = f'{x + offsets[2][1]}_{y + offsets[2][2]}_{direction}.html'
                else:
                    forwardURL = '#'

                # Create the .html file:
                htmlFile.write(f'''
<!-- This line is an HTML comment, which is ignored by the browser. -->
<!-- HTML tags are between angle brackets < and >. The <html>, <head>,
     <title>, and <body> tags are used in all web pages. They are paired
     with closing tags like </html>. -->
<html>
<head>
    <title>{filename}</title>
</head>
<body>
    <!-- <center> puts things in the center of the page. -->
    <center>
    <!-- <img> displays an image, while <br /> adds a break return.
         The <br> tag has no paired closing tag, which is why it's
         written as <br /> -->
    <img src="../maze_html_images/{wallImageFilename}" /><br />

    Facing: {direction}<br />

    <!-- Everything between <a> and </a> is a clickable link. We make the
         arrow images clickable links. -->
    <a href="{turnLeftURL}"><img src="../maze_html_images/turn_left.png" /></a>
    <a href="{forwardURL}"><img src="../maze_html_images/forward.png" /></a>
    <a href="{turnRightURL}"><img src="../maze_html_images/turn_right.png" /></a>
    </center>
</body>
</html>
''')
            # After writing the html file, see if this is the starting position:
            if (startx, starty, direction) == (x, y, NORTH):
                shutil.copy(htmlFilename, f'{outputFolderName}/index.html')
                numFilesWritten += 1



print(f'Done. {numFilesWritten} files written.')


# TODO - add exit logic, obfuscate the URLs